from PIL import Image, ImageFont, ImageDraw, ImageEnhance
import numpy
import random
import operator
import imageio
from fpdf import FPDF

class pcp_core:

  #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~+
  # NOTES
  #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~+
  # The coordinate (0,0) of the image is the top left corner!
  #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~+

  pieces = []
  cards = []

  # overall resolution. I want at least 50 pixel / cm to give it a good high res
  # as such 5 pixel / mm
  r_ppmm = 6

  # the final image is 900 mm x 600 mm
  final_h_mm = 600
  final_w_mm = 900
  final_h_p = r_ppmm * final_h_mm
  final_w_p = r_ppmm * final_w_mm

  # final image will have 18 x 12 pieces
  pieces_h_count = 12
  pieces_w_count = 18

  # pieces height and width
  pieces_h_mm = final_h_mm / pieces_h_count
  pieces_w_mm = final_w_mm / pieces_w_count
  pieces_h_p = pieces_h_mm * r_ppmm
  pieces_w_p = pieces_w_mm * r_ppmm

  # a card is 150 mm x 100 mm 
  cards_h_mm = 100
  cards_w_mm = 150
  cards_h_p = cards_h_mm * r_ppmm
  cards_w_p = cards_w_mm * r_ppmm

  # pieces per card
  cards_h_pieces = cards_h_mm / pieces_h_mm
  cards_w_pieces = cards_w_mm / pieces_w_mm
  cards_pieces_count = cards_h_pieces * cards_w_pieces

  # border around each image to ensure clean cuts after print
  border_p = 25 

  # sequence size factor
  sequence_size_divider = 10 

  # the general background of each post card
  gbg = -1

  # address where the postcard shall go, 3 lines
  address = ("First Name", "Road & Number", "Postcode, City")

  # overall message
  # one character per card
  overall_message = "Message which will be split between the cards and it actually has to be 216 characters long to work to have one character per puzzle piece. Otherwise it will just run into a stupid out of bounds exception :):):):):):"

  font_name = -1
  tfont = -1
  font_spacer = -1

  # project name
  project_name = -1

  placer = -1

##############################

  class general_bg:
    over = -1

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~+
    # postcards back color
    background_color = (255,0,0,0)# "white" # needs to be set to (255,0,0,0) for transparent
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~+

    img = -1
    draw = -1

    def __init__(self,over):
      self.over = over

    def do(self):
      print "Generating general back side started"
      self.img = Image.new( 'RGBA', (self.over.cards_w_p,self.over.cards_h_p), self.background_color)
      self.draw = ImageDraw.Draw(self.img)
      self.draw_general_bg()
      print "Generating general back side finished"

    def draw_general_bg(self): 
      self.draw_grid()
      self.draw_address_lines()
      self.draw_center_half()

    def draw_grid(self):
      #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~+
      # settings for grid
      grid_color = (232,232,232) #"black" # light grey
      grid_w_mm = 0.8 # in mm
      #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~+
      grid_w_p = self.over.r_ppmm * grid_w_mm

      # actual drawing
      # horizontal line
      self.draw.rectangle( ( (0, ((self.over.cards_h_p / 2) - (grid_w_p / 2))), (self.over.cards_w_p, ((self.over.cards_h_p / 2) + (grid_w_p / 2))) ), fill=grid_color)
      # vertical 1
      self.draw.rectangle( ( ( ((self.over.cards_w_p / 3) - (grid_w_p / 2)), 0), ( ((self.over.cards_w_p / 3) + (grid_w_p / 2)), self.over.cards_h_p) ), fill=grid_color)
      # vertical 2
      self.draw.rectangle( ( ( (((self.over.cards_w_p / 3)*2) - (grid_w_p / 2)), 0), ( (((self.over.cards_w_p / 3)*2) + (grid_w_p / 2)), self.over.cards_h_p) ), fill=grid_color)

    def draw_address_lines(self):
      #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~+
      # settings for address
      address_line_distance_mm = 8
      address_line_w_mm = 0.5 # in mm
      address_font_name = "Pillow/Tests/fonts/FreeMono.ttf"
      font_color = "black"
      address_spacer_left_mm = 90
      address_spacer_top_mm = 70
      address_spacer_right_mm = 10
      #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~+
      address_line_w_p = self.over.r_ppmm * address_line_w_mm
      address_line_distance_p = self.over.r_ppmm * address_line_distance_mm
      address_spacer_left_p = self.over.r_ppmm * address_spacer_left_mm
      address_spacer_top_p = self.over.r_ppmm * address_spacer_top_mm
      address_spacer_right_p = self.over.r_ppmm * address_spacer_right_mm
      address_end_right_p = self.over.cards_w_p - address_spacer_right_p

      for i in (0,1,2):
        self.draw.text((address_spacer_left_p,(address_spacer_top_p+(i*address_line_distance_p)-self.over.auto_font(self.over.address[i], address_font_name, self.over.cards_w_p-address_spacer_left_p-address_spacer_right_p,address_line_distance_p)[1]-address_line_w_p)), self.over.address[i], font=self.over.auto_font(self.over.address[i], address_font_name, self.over.cards_w_p-address_spacer_left_p-address_spacer_right_p,address_line_distance_p)[0], fill="black")
        self.draw.rectangle( ( (address_spacer_left_p, (address_spacer_top_p + (i*address_line_distance_p) - (address_line_w_p/2)) ), (address_end_right_p, (address_spacer_top_p + (i*address_line_distance_p) + (address_line_w_p/2)) ) ), fill="black")

    def draw_center_half(self):
      #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~+
      # settings for center line
      center_line_width_mm = 1
      gaps_mm = 5
      #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~+
      center_line_width_p = self.over.r_ppmm * center_line_width_mm
      gaps_p = self.over.r_ppmm * gaps_mm

      self.draw.rectangle( ( (((self.over.cards_w_p/2)-(center_line_width_p/2)),gaps_p), (((self.over.cards_w_p/2) + (center_line_width_p/2)),((self.over.cards_h_p/2)-gaps_p))), fill = (176,176,176))
      self.draw.rectangle( ( (((self.over.cards_w_p/2)-(center_line_width_p/2)),((self.over.cards_h_p/2)+gaps_p)), (((self.over.cards_w_p/2) + (center_line_width_p/2)),(self.over.cards_h_p-gaps_p))), fill = (176,176,176))

####################################################

  class piece:
    over = -1
    position = -1
    image = -1
#    imageb = -1
    back = -1
    order = -1
    orientation = -1
    msg = -1

    def __init__(self, position, image, msg, over):
      self.over = over
      self.position = position
      self.image = image
#      self.imageb = Image.new( 'RGBA', (self.over.pieces_w_p + 2* self.over.border_p,self.over.pieces_h_p + 2* self.over.border_p), (255,0,0,0))
#      self.imageb.paste(image,(0,0))
      self.orientation = random.randint(0,3)
      self.msg = msg

    def make_bg(self,card,c):
      #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~+
      # settings for markings
      caption = "C" + str(card) + "(" + str(self.position[0]) + "|" + str(self.position[1]) + ")" + str(c) + self.msg
      #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~+

      self.back = Image.new( 'RGBA', (self.over.pieces_w_p,self.over.pieces_h_p), (255,0,0,0))
      draw = ImageDraw.Draw(self.back)
      t_len = len(caption)
      draw.text((((self.over.pieces_w_p-(t_len*(self.over.font_spacer/2)))/2),self.over.pieces_h_p-(self.over.font_spacer+8)), caption, font=self.over.tfont, fill="black")

    def p(self):
      print str(self.position) + " @ " + str(self.order) + ", " + str(self.orientation) + ": " + str(self.image)

    def get_image(self):
      t = self.image.crop((self.over.border_p, self.over.border_p, self.over.pieces_w_p + 2* self.over.border_p - self.over.border_p, self.over.pieces_w_p + 2* self.over.border_p - self.over.border_p))
      if self.orientation == 1:
        return t.rotate(90)
      elif self.orientation == 2:
        return t.rotate(180)
      elif self.orientation == 3:
        return t.rotate(270)
      else:
        return t

    def get_image_border(self, c):
      if self.orientation == 1:
        return self.border_cut(self.image.rotate(90),c)
      elif self.orientation == 2:
        return self.border_cut(self.image.rotate(180),c)
      elif self.orientation == 3:
        return self.border_cut(self.image.rotate(270),c)
      else:
        return self.border_cut(self.image,c)

    def border_cut(self, image, c):
      # top left image
      if c == 0:
        return image.crop((0, 0, self.over.border_p + self.over.pieces_w_p, self.over.border_p + self.over.pieces_h_p))
      # top middle image
      if c == 1:
        return image.crop((self.over.border_p, 0, self.over.border_p + self.over.pieces_w_p, self.over.border_p + self.over.pieces_h_p))
      # top right image
      if c == 2:
        return image.crop((self.over.border_p, 0, self.over.border_p + self.over.pieces_w_p + self.over.border_p, self.over.border_p + self.over.pieces_h_p))
      # bottom left image
      if c == 3:
        return image.crop((0, self.over.border_p, self.over.border_p + self.over.pieces_w_p, self.over.border_p + self.over.pieces_h_p + self.over.border_p))
      # bottom middle image
      if c == 4:
        return image.crop((self.over.border_p, self.over.border_p, self.over.border_p + self.over.pieces_w_p, self.over.border_p + self.over.pieces_h_p + self.over.border_p))
      # bottom right image
      if c == 5:
        return image.crop((self.over.border_p, self.over.border_p, self.over.border_p + self.over.pieces_w_p + self.over.border_p, self.over.border_p + self.over.pieces_w_p + self.over.border_p))


    def get_back(self):
      if self.orientation == 1:
        return self.back.rotate(90)
      elif self.orientation == 2:
        return self.back.rotate(180)
      elif self.orientation == 3:
        return self.back.rotate(270)
      else:
        return self.back

######################################################################

  class card:
    pieces = []
    front = -1
    front_over = -1
    back = -1
    no = -1
    over = -1

    def __init__(self, no, images, over):
      self.no = no
      self.pieces = images
      self.over = over
      print "Card " + str(no) + " has been created"

    def save(self):
      filename = "pc_" + str(self.no)
      tf=self.front.convert("RGB")
      tb=self.back.convert("RGB")
      tf.save(project_name + "_" + filename+"_front.png")
      tb.save(project_name + "_" + filename+"_back.png")

    def piece_bgs(self):
      c = 0
      for piece in self.pieces:
        piece.make_bg(self.no,c)
        c+=1

    def fix_orientation(self):
      while self.pieces[4].orientation in [1]:
        self.pieces[4].orientation = random.randint(0,3)  
      while self.pieces[3].orientation in [1,3]:
        self.pieces[3].orientation = random.randint(0,3)

    def p(self):
      print "Card " + str(self.no) + str(self.pieces)

    def generate(self):
      self.piece_bgs()
      self.fix_orientation()
      self.merge_back()
      self.merge_front()
      self.merge_front_over()
      print "Card " + str(self.no) + " has been generated"

    def merge_back(self):
      self.back = Image.new('RGBA', (self.over.cards_w_p, self.over.cards_h_p), "white")
      c = 0
      for h in range(0,2):
        for v in range(2,-1,-1): # count bardwards 2,1,0 to reverse images as they'll be on the backside!
          self.back.paste(self.pieces[c].get_back(), (self.over.pieces_w_p*v,self.over.pieces_h_p*h),self.pieces[c].get_back())
          c += 1
      self.back.paste(self.over.gbg.img, (0,0), self.over.gbg.img)

    def merge_front(self):
      self.front = Image.new('RGBA', (self.over.cards_w_p, self.over.cards_h_p), "green")
      c = 0
      for h in range(0,2):
        for v in range(0,3):
          self.front.paste(self.pieces[c].get_image(), (self.over.pieces_w_p*v,self.over.pieces_h_p*h))
          c += 1

    def merge_front_over(self):
      self.front_over = Image.new('RGBA', (self.over.cards_w_p + 2 * self.over.border_p, self.over.cards_h_p + 2 * self.over.border_p), "green")


      self.front_over.paste(self.pieces[0].get_image_border(0), (0,0))
      self.front_over.paste(self.pieces[1].get_image_border(1), (self.over.border_p + self.over.pieces_w_p,0))
      self.front_over.paste(self.pieces[2].get_image_border(2), (self.over.border_p + 2 * self.over.pieces_w_p,0))
      self.front_over.paste(self.pieces[3].get_image_border(3), (0,self.over.border_p + self.over.pieces_w_p))
      self.front_over.paste(self.pieces[4].get_image_border(4), (self.over.border_p + self.over.pieces_w_p, self.over.border_p + self.over.pieces_w_p))
      self.front_over.paste(self.pieces[5].get_image_border(5), (self.over.border_p + 2 * self.over.pieces_w_p,self.over.border_p+self.over.pieces_w_p))



###########################################################

  class test_img:

    font_size = 14
    address_font = ImageFont.truetype('Pillow/Tests/fonts/FreeMono.ttf', font_size)
    font_color = "black"

    img = -1
    draw = -1

    def __init__(self,over):
      self.img = Image.new( 'RGBA', (over.final_w_p+(over.border_p*2),over.final_h_p+(over.border_p*2)), "white")
      self.draw = ImageDraw.Draw(self.img)

      c = 0
      print over.pieces_h_count
      print over.pieces_w_count
      for cy in range (0,over.pieces_h_count):
        for cx in range (0,over.pieces_w_count):
          text = "c:" + str(c) + "@(" + str(cx) + "|" + str(cy) + ")"
          x = over.border_p + ( cx * over.pieces_w_p) + (over.pieces_w_p/2) - ((len(text)*self.font_size/2))
          y = over.border_p + (cy * over.pieces_h_p) + (over.pieces_h_p/2)
          self.draw.text((x,y), text, font=self.address_font, fill="black")
          c+=1

      for cx in range (0,over.pieces_w_count):
        self.draw.line((over.border_p + over.pieces_w_p * cx, 0, over.border_p + over.pieces_w_p * cx, over.final_h_p+(over.border_p*2)), "green")

      for cy in range (0,over.pieces_h_count):
        self.draw.line((0, over.border_p + over.pieces_h_p * cy, over.final_w_p + (over.border_p*2), over.border_p + (over.pieces_h_p * cy)), "green")

      self.draw.line((0, over.border_p, over.final_w_p + (over.border_p*2), over.border_p), "red")
      self.draw.line((0, over.final_h_p+(over.border_p*2)-over.border_p, over.final_w_p + (over.border_p*2), over.final_h_p+(over.border_p*2)-over.border_p), "red")
      self.draw.line((over.border_p, 0, over.border_p,over.final_h_p+(over.border_p*2)), "red")
      self.draw.line((over.final_w_p+(over.border_p*2)-over.border_p, 0, over.final_w_p+(over.border_p*2)-over.border_p,over.final_h_p+(over.border_p*2)), "red")

  
    def save(self):
      self.img.convert("RGB").save("eval.jpg")


##############################

  def __init__(self, project_name):
    self.project_name = project_name
    self.gbg = self.general_bg(self)
    self.font_name = "Pillow/Tests/fonts/FreeMono.ttf"
    self.tfont = self.auto_font("Cnn(ab|ab)n", self.font_name, 200, (self.r_ppmm*4))[0]
    self.font_spacer = self.auto_font("Cnn(ab|ab)n", self.font_name, 100, (self.r_ppmm*2))[1]*2

##############################

  def generate_marker(self):
    self.placer = Image.new('RGBA', (6 * self.r_ppmm, 6 * self.r_ppmm), "white")
    dmark = ImageDraw.Draw(self.placer)
    dmark.ellipse((1*self.r_ppmm, 1*self.r_ppmm, 5*self.r_ppmm, 5*self.r_ppmm), fill="black")
    dmark.ellipse((1.5*self.r_ppmm, 1.5*self.r_ppmm, 4.5*self.r_ppmm, 4.5*self.r_ppmm), fill="white")

    dmark.line(((6*self.r_ppmm)/2,0,(6*self.r_ppmm)/2,6*self.r_ppmm), fill="black")
    dmark.line(((0,(6*self.r_ppmm)/2,6*self.r_ppmm,(6*self.r_ppmm)/2)), fill="black")

    dmark.ellipse((2*self.r_ppmm, 2*self.r_ppmm, 4*self.r_ppmm, 4*self.r_ppmm), fill="black")

    dmark.line(((6*self.r_ppmm)/2, 2*self.r_ppmm, (6*self.r_ppmm)/2, 4*self.r_ppmm), fill="white")
    dmark.line(((2*self.r_ppmm ,(6*self.r_ppmm)/2, 4*self.r_ppmm, (6*self.r_ppmm)/2)), fill="white")


  def generate_10x15_border_pdf(self):
    print "Generating 15x10 pdf"
    
    border = self.r_ppmm * 10
    page_w_p = (self.r_ppmm * 150) + (2 * border)
    page_h_p = (self.r_ppmm * 100) + (2 * border)

    pages = []
    for i in range (0,36):
      page = Image.new('RGBA', (page_w_p, page_h_p), "white")
      page.paste(self.cards[i].front,(border,border))
      pages += [page]
      page = Image.new('RGBA', (page_w_p, page_h_p), "white")
      page.paste(self.cards[i].back,(border,border))
      pages += [page]

    print "Created " + str(len(pages)) + " 15x10 pages, combining now"

    pdf = FPDF(unit = "mm", format = (170, 120))
    c = 0
    for page in pages:
      page.convert("RGB").save("pdfwork/" + self.project_name + "_10x15_border_" + str(c)+".jpg", format='JPEG', subsampling=0, quality=100)
      pdf.add_page()
      pdf.image("pdfwork/" + self.project_name + "_10x15_border_" + str(c)+".jpg",0,0,170,120)
      c+=1
    pdf.output(self.project_name + "_10x15_border.pdf", "F")

    # cut marks & placer
    for page in pages:
      draw = draw = ImageDraw.Draw(page)
      #top left
      draw.line((self.r_ppmm*10, 2*self.r_ppmm, self.r_ppmm*10, 7*self.r_ppmm), fill=(0,0,0))
      draw.line((2*self.r_ppmm, self.r_ppmm*10, 7*self.r_ppmm, self.r_ppmm*10), fill=(0,0,0))

      #top right
      draw.line((page_w_p - (self.r_ppmm * 10), 2*self.r_ppmm, page_w_p - (self.r_ppmm * 10), 7*self.r_ppmm), fill=(0,0,0))
      draw.line((page_w_p - (7*self.r_ppmm), self.r_ppmm*10, page_w_p - (2*self.r_ppmm), self.r_ppmm*10), fill=(0,0,0))

      #bottom right
      draw.line((page_w_p - (self.r_ppmm*10), page_h_p - (7*self.r_ppmm), page_w_p - (self.r_ppmm*10), page_h_p - (2*self.r_ppmm)), fill=(0,0,0))
      draw.line((page_w_p - (7*self.r_ppmm), page_h_p - (10*self.r_ppmm), page_w_p - (2*self.r_ppmm), page_h_p - (10*self.r_ppmm)), fill=(0,0,0))

      #bottom left
      draw.line((self.r_ppmm*10, page_h_p - (7*self.r_ppmm), self.r_ppmm*10, page_h_p - (2*self.r_ppmm)), fill=(0,0,0))
      draw.line((2*self.r_ppmm, page_h_p - (10*self.r_ppmm), 7*self.r_ppmm, page_h_p - (10*self.r_ppmm)), fill=(0,0,0))

      page.paste(self.placer, (((page_w_p/2)-3*self.r_ppmm),(2*self.r_ppmm)),self.placer)
      page.paste(self.placer, (((page_w_p/2)-3*self.r_ppmm),page_h_p-(2*self.r_ppmm)-(6*self.r_ppmm)),self.placer)


    pdf = FPDF(unit = "mm", format = (170, 120))
    c = 0
    for page in pages:
      page.convert("RGB").save("pdfwork/" + self.project_name + "_10x15_border_marker" + str(c)+".jpg", format='JPEG', subsampling=0, quality=100)
      pdf.add_page()
      pdf.image("pdfwork/" + self.project_name + "_10x15_border_marker" + str(c)+".jpg",0,0,170,120)
      c+=1
    pdf.output(self.project_name + "_10x15_border_marker.pdf", "F")
           
  def generate_10x15_over_pdf(self):
    print "Generating 15x10 pdf"
    
    border = self.r_ppmm * 10
    page_w_p = (self.r_ppmm * 150) + (2 * border)
    page_h_p = (self.r_ppmm * 100) + (2 * border)

    pages = []
    for i in range (0,36):
      page = Image.new('RGBA', (page_w_p, page_h_p), "white")
      page.paste(self.cards[i].front_over,(border-self.border_p,border-self.border_p))
      pages += [page]
      page = Image.new('RGBA', (page_w_p, page_h_p), "white")
      page.paste(self.cards[i].back,(border,border))
      pages += [page]

    print "Created " + str(len(pages)) + " 15x10 pages, combining now"

    pdf = FPDF(unit = "mm", format = (170, 120))
    c = 0
    for page in pages:
      page.convert("RGB").save("pdfwork/" + self.project_name + "_10x15_border_" + str(c)+".jpg", format='JPEG', subsampling=0, quality=100)
      pdf.add_page()
      pdf.image("pdfwork/" + self.project_name + "_10x15_border_" + str(c)+".jpg",0,0,170,120)
      c+=1
    pdf.output(self.project_name + "_10x15_border_over.pdf", "F")

    # place cut marks & placer
    for page in pages:
      draw = draw = ImageDraw.Draw(page)
      #top left
      draw.line((self.r_ppmm*10, 2*self.r_ppmm, self.r_ppmm*10, 7*self.r_ppmm), fill=(0,0,0))
      draw.line((2*self.r_ppmm, self.r_ppmm*10, 7*self.r_ppmm, self.r_ppmm*10), fill=(0,0,0))

      #top right
      draw.line((page_w_p - (self.r_ppmm * 10), 2*self.r_ppmm, page_w_p - (self.r_ppmm * 10), 7*self.r_ppmm), fill=(0,0,0))
      draw.line((page_w_p - (7*self.r_ppmm), self.r_ppmm*10, page_w_p - (2*self.r_ppmm), self.r_ppmm*10), fill=(0,0,0))

      #bottom right
      draw.line((page_w_p - (self.r_ppmm*10), page_h_p - (7*self.r_ppmm), page_w_p - (self.r_ppmm*10), page_h_p - (2*self.r_ppmm)), fill=(0,0,0))
      draw.line((page_w_p - (7*self.r_ppmm), page_h_p - (10*self.r_ppmm), page_w_p - (2*self.r_ppmm), page_h_p - (10*self.r_ppmm)), fill=(0,0,0))

      #bottom left
      draw.line((self.r_ppmm*10, page_h_p - (7*self.r_ppmm), self.r_ppmm*10, page_h_p - (2*self.r_ppmm)), fill=(0,0,0))
      draw.line((2*self.r_ppmm, page_h_p - (10*self.r_ppmm), 7*self.r_ppmm, page_h_p - (10*self.r_ppmm)), fill=(0,0,0))

      page.paste(self.placer, (((page_w_p/2)-3*self.r_ppmm),(2*self.r_ppmm)),self.placer)
      page.paste(self.placer, (((page_w_p/2)-3*self.r_ppmm),page_h_p-(2*self.r_ppmm)-(6*self.r_ppmm)),self.placer)


    pdf = FPDF(unit = "mm", format = (170, 120))
    c = 0
    for page in pages:
      page.convert("RGB").save("pdfwork/" + self.project_name + "_10x15_border_marker" + str(c)+".jpg", format='JPEG', subsampling=0, quality=100)
      pdf.add_page()
      pdf.image("pdfwork/" + self.project_name + "_10x15_border_marker" + str(c)+".jpg",0,0,170,120)
      c+=1
    pdf.output(self.project_name + "_10x15_border_over_marker.pdf", "F")


  def generate_a4_pdf(self):
    print "Generating A4 PDF"
    # A4 measurements
    a4_w_mm = 210
    a4_h_mm = 297
    a4_w_p = self.r_ppmm * a4_w_mm
    a4_h_p = self.r_ppmm * a4_h_mm

    # equal distances between images and borders
    spacer_w_p = (a4_w_p - self.cards_w_p) / 2
    spacer_h_p = (a4_h_p - (2 * self.cards_h_p)) / 3

    pages = []

    for i in range (0,36,2):
      pagef = Image.new( 'RGBA', (a4_w_p, a4_h_p), "white")
      pagef.paste(self.cards[i].front,(spacer_w_p,spacer_h_p))
      pagef.paste(self.cards[i+1].front,(spacer_w_p,spacer_h_p + self.cards_h_p + spacer_h_p))
      pages += [pagef]
      print "Page " + str(i) + " front done"
      pageb = Image.new( 'RGBA', (a4_w_p, a4_h_p), "white")
      pageb.paste(self.cards[i].back,(spacer_w_p,spacer_h_p))
      pageb.paste(self.cards[i+1].back,(spacer_w_p,spacer_h_p + self.cards_h_p + spacer_h_p))
      pages += [pageb]
      print "Page " + str(i) + " back done"

    print "Created " + str(len(pages)) + " A4 pages, combining now"
    pdf = FPDF()
    c = 0
    for page in pages:
      page.convert("RGB").save("pdfwork/" + self.project_name + "_a4_" + str(c)+".jpg", format='JPEG', subsampling=0, quality=100)
      pdf.add_page()
      pdf.image("pdfwork/" + self.project_name + "_a4_" + str(c)+".jpg",0,0,a4_w_mm,a4_h_mm)
      c+=1
    pdf.output(self.project_name + "_a4.pdf", "F")

    print "Generating A4 PDF done"
    

  def create_sequence_gif(self):
    print "Starting to create sequence gif"
    steps = []
    work_image = Image.new( 'RGBA', (self.final_w_p,self.final_h_p), "white")
    c = 0
    for card in self.cards:
      for piece in card.pieces:
        work_image.paste(piece.image, (piece.position[0] * self.pieces_w_p,piece.position[1] * self.pieces_h_p))
      steps += [numpy.array(work_image.resize((self.final_w_p/self.sequence_size_divider,self.final_h_p/self.sequence_size_divider)))]
      print "Step " + str(c) + " created"
      c += 1

    kargs = { 'duration': 0.5 }
    imageio.mimsave(self.project_name + "_build_sequence.gif", steps, 'GIF', **kargs)
    print "Finished creating sequence gif"

  def auto_font(self, text, font_name, max_width, max_height):
    fontsize_count = 1
    font = ImageFont.truetype(font_name, fontsize_count)
    while (font.getsize(text)[0] < max_width) and (font.getsize(text)[1] < max_height):
      fontsize_count += 1
      font = ImageFont.truetype(font_name, fontsize_count)

    fontsize_count -= 1
    return [ImageFont.truetype(font_name, fontsize_count),font.getsize(text)[1]]

  def slice_image(self, image):
    print "Image slicing started"
    # load picture
    image = Image.open(image)
    width, height = image.size
    print "Loaded image is " + str(width) + " x " + str(height)
    if (width < self.final_w_p + 2*self.border_p) or (height < self.final_h_p + 2*self.border_p):
      print "Target size is " + str(self.final_w_p) + " x " + str(self.final_h_p) + ": resizing!"
      image = image.resize((self.final_w_p + 2*self.border_p,self.final_h_p + 2*self.border_p))

    # the actual slicing and creation of pieces
    print "Now slicing into " + str(self.pieces_h_count * self.pieces_w_count) + " pieces with size of: " + str(self.pieces_w_p) + " x " + str(self.pieces_h_p)
    c = 0
    for ch in range(0,self.pieces_h_count):
      for cw in range(0,self.pieces_w_count):
        x1 = self.border_p + (cw * self.pieces_w_p) - self.border_p
        y1 = self.border_p + (ch * self.pieces_h_p) - self.border_p
        x2 = self.border_p + self.pieces_w_p + (cw * self.pieces_w_p) + self.border_p
        y2 = self.border_p + self.pieces_h_p + (ch * self.pieces_h_p) + self.border_p
        p = image.crop((x1,y1,x2,y2))
        self.pieces += [self.piece((cw,ch),p,self.overall_message[c],self)]
        c += 1

    print "Image slicing finished. " + str(len(self.pieces)) + " pieces have been created"

  def set_order(self, image):
    print "Adding order started"
    image = Image.open(image)
    image = image.convert('RGB')

    for ch in range(0,self.pieces_h_count):
      for cw in range(0,self.pieces_w_count):
        r,g,b = image.getpixel((cw,ch))
        self.add_order((cw,ch),b)
    print "Adding order finished"

  def add_order(self, pos, order):
    for piece in self.pieces:
      if piece.position == pos:
        piece.order = order

  def generate_cards(self):
    print "Generating cards started"
    self.gbg.do()
    random.shuffle(self.pieces)
    print "Pieces have been shuffled"
    self.pieces.sort(key=operator.attrgetter('order'))
    print "Pieces have been sorted by order"

    c = 0
    for i in range(0,36):
      self.cards += [self.card(i,[self.pieces[c+0],self.pieces[c+1],self.pieces[c+2], self.pieces[c+3],self.pieces[c+4],self.pieces[c+5]],self)]
      c += 6
    print "Cards have been created"
    for card in self.cards:
      card.generate()

  def save(self):
    print "Saving cards"
    for card in self.cards:
      card.save()

  def do_it(self, image, order_map):
    print "Starting to generate postcards"
    self.generate_marker()
    self.slice_image(image)
    self.set_order(order_map)
    self.generate_cards()
    self.create_sequence_gif()
    self.generate_a4_pdf()
    self.generate_10x15_border_pdf()
    self.generate_10x15_over_pdf()
    #self.save()
    print "DONE! :)"
    
  def create_test_img(self):
    t = self.test_img(self)
    t.save()
